# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
pyspades - Ace of Spades network protocol implementations
"""

__author__ = 'Mathias Kaerlev'
__copyright__ = 'Copyright 2011, Mathias Kaerlev'
__credits__ = ['Mathias "Mat^2" Kaerlev']
__version__ = '0.1'
__email__ = 'matpow2@gmail.com'

__all__ = ['vxl', 'bytes', 'contained', 'collision', 'common', 'constants',
           'loaders', 'master', 'packet', 'protocol', 'server', 'tools',
           'types']
