#ifndef CONSTANTS_C_H
#define CONSTANTS_C_H

#define PI 3.141592653589793f
#define VXL_MAX_SIZE 512 // maximum .VXL dimensions in both x & y direction
#define VSID VXL_MAX_SIZE
#define VXL_MAX_SIZEM (VXL_MAX_SIZE - 1)
#define VSIDM VXL_MAX_SIZEM
#define VSIDSQ (VXL_MAX_SIZE * VXL_MAX_SIZE)

#endif
